#include "globals.h"
#include <time.h>

//16mhz
#define DBG_MCU_SPEED 16000000

#ifdef LOG

static MUTEX_DECL(m1);

#define CHANNEL_CONTROL 0

#define PACKET_SET_CHANNEL_NAME	(0 << 5)
#define PACKET_TEXT 			(1 << 5)
#define PACKET_BINARY_SEND		(2 << 5)
#define PACKET_BINARY_RECEIVE	(3 << 5)

static u8 swoTmp[4];
static void swoWrite(int channel, uint8_t *data, int length);
#define swoWrite8(channel, data) {swoTmp[0]=data; swoWrite(channel, swoTmp, 1);}
#define swoWrite16(channel, data) {swoTmp[0]=data>>8; swoTmp[1]=data; swoWrite(channel, swoTmp, 2);}
#define swoWrite32(channel, data) {swoTmp[0]=data>>24; swoTmp[1]=data>>16; swoTmp[2]=data>>8; swoTmp[3]=data; swoWrite(channel, swoTmp, 4);}

static int printfChannel=1;

#define startPacket(channel, type) swoWrite8(CHANNEL_CONTROL,type|channel);

void dbgInit(void)
{
	chMtxInit(&m1);
	pinSetMode(&PIN_TRIGGER,PAL_MODE_OUTPUT_PUSHPULL);
	u32 SWOSpeed = 1000000; //1000kbps
	u32 SWOPrescaler = (DBG_MCU_SPEED / SWOSpeed) - 1; // SWOSpeed in Hz

	*((volatile unsigned *)0xE000EDFC) = 0x01000000;   // "Debug Exception and Monitor Control Register (DEMCR)"
	*((volatile unsigned *)0xE0042004) = 0x00000027;	  // Enabling TRACE_IOEN, DBG_STANDBY, DBG_STOP, DBG_SLEEP
	*((volatile unsigned *)0xE00400F0) = 0x00000002;   // "Selected PIN Protocol Register": Select which protocol to use for trace output (2: SWO)
	*((volatile unsigned *)0xE0040010) = SWOPrescaler; // "Async Clock Prescaler Register". Scale the baud rate of the asynchronous output
	*((volatile unsigned *)0xE0000FB0) = 0xC5ACCE55;   // ITM Lock Access Register, C5ACCE55 enables more write access to Control Register 0xE00 :: 0xFFC
	*((volatile unsigned *)0xE0000E80) = 0x0001000D;   // ITM Trace Control Register
	*((volatile unsigned *)0xE0000E40) = 0xFFFFFFFF;   // ITM Trace Privilege Register
	*((volatile unsigned *)0xE0000E00) = 0xFFFFFFFF;   // ITM Trace Enable Register. Enabled tracing on stimulus ports. One bit per stimulus port.
	*((volatile unsigned *)0xE0001000) = 0x400003FE;   // DWT_CTRL
	*((volatile unsigned *)0xE0040304) = 0x00000100;   // Formatter and Flush Control Register
}

void dbgSetChannelName(int channel, char *name)
{
	chMtxLock(&m1);

	startPacket(channel,PACKET_SET_CHANNEL_NAME);
	swoWrite(channel,(u8 *)name,strlen(name));
	startPacket(channel,PACKET_TEXT);

	chMtxUnlock();
}

static void printfch(char c)
{
	swoWrite8(printfChannel,c);
}


//packet functions
inline void dbgStartS(int channel)
{
	startPacket(channel,PACKET_TEXT);
}

inline void dbgStartW(int channel)
{
	startPacket(channel,PACKET_BINARY_SEND);
}

inline void dbgStartR(int channel)
{
	startPacket(channel,PACKET_BINARY_RECEIVE);
}

inline void dbgb(int channel, u8 b)
{
	swoWrite8(channel,b);
}

inline void dbgd(int channel, u8 *buffer, int length)
{
	chMtxLock(&m1);

	swoWrite(channel,buffer,length);

	chMtxUnlock();
}

inline void dbgEnd(int channel)
{
	startPacket(channel,PACKET_TEXT);
}


//string functions
void dbgf(int channel, char *fmt, ...)
{
	chMtxLock(&m1);

	va_list va;

	va_start(va,fmt);
	printfChannel=channel;
	dbgStartS(channel);
	chvsprintf(0,fmt,va,printfch);
	dbgEnd(channel);
	va_end(va);

	chMtxUnlock();
}

void dbgStr(int channel, char *string)
{
	chMtxLock(&m1);

	dbgStartS(channel);
	swoWrite(channel,(u8 *)string,strlen(string));
	dbgEnd(channel);

	chMtxUnlock();
}

void dbgDataR(int channel, void *buffer, int length)
{
	chMtxLock(&m1);

	dbgStartR(channel);
	swoWrite(channel,buffer,length);
	dbgEnd(channel);

	chMtxUnlock();
}

void dbgDataW(int channel, void *buffer, int length)
{
	chMtxLock(&m1);

	dbgStartW(channel);
	swoWrite(channel,buffer,length);
	dbgEnd(channel);

	chMtxUnlock();
}


//printf and the rest
void chprintf(const char *fmt, ...)
{
	va_list va;

	va_start(va,fmt);
	chvsprintf(0,fmt,va,printfch);
	va_end(va);
}

#endif

#define MAX_FILLER 11

char *ltoa(char *p, long num, unsigned radix)
{
	int i;
	char *q;

	q=p+MAX_FILLER;
	do
	{
		i=(int)(num%radix);
		i+='0';
		if (i>'9')
			i+='A'-'0'-10;
		*--q=i;
	} while ((num/=radix)!=0);

	i=(int)(p+MAX_FILLER-q);
	do
		*p++=*q++;
	while (--i);

	return p;
}

/**
 * @brief   System formatted output function.
 * @details This function implements a minimal @p printf() like functionality
 *          with output on a @p BaseChannel.
 *          The general parameters format is: %[.][width|*][l|L]p.
 *          The following parameter types (p) are supported:
 *          - <b>x</b> hexadecimal integer.
 *          - <b>X</b> hexadecimal long.
 *          - <b>o</b> octal integer.
 *          - <b>O</b> octal long.
 *          - <b>d</b> decimal signed integer.
 *          - <b>D</b> decimal signed long.
 *          - <b>u</b> decimal unsigned integer.
 *          - <b>U</b> decimal unsigned long.
 *          - <b>c</b> character.
 *          - <b>s</b> string.
 *          .
 * @note    Floating point types are not implemented, this function is meant
 *          as a system utility and not a full implementation.
 *
 * @param[in] chp       pointer to a @p BaseChannel implementing object
 * @param[in] fmt       formatting string
 */
void chvsprintf(char *dest, const char *fmt, va_list va, printfch_t printfch)
{
	char buf[MAX_FILLER+1];
	char *p, *s, c, filler;
	int i, n, width;
	bool_t lflag, ljust;
	long l;

	while (TRUE)
	{
		c=*fmt++;
		if (c==0)
		{
			va_end(va);
			if(dest)
				*dest=0;
			return;
		}
		if (c!='%')
		{
			if(dest)
				*dest++=c;
			else
				printfch(c);
			continue;
		}
		p=buf;
		s=buf;
		ljust=FALSE;
		if (*fmt=='-')
		{
			fmt++;
			ljust=TRUE;
		}
		filler=' ';
		if (*fmt=='.')
		{
			fmt++;
			filler='0';
		}
		for (width=0;;)
		{
			c=*fmt++;
			if (c>='0'&&c<='9')
				c-='0';
			else
				if (c=='*')
					c=va_arg(va, int);
				else
					break;
			width=width*10+c;
		}
		n=0;
		if (c=='.')
		{
			for (;;)
			{
				c=*fmt++;
				if (c>='0'&&c<='9')
					c-='0';
				else
					if (c=='*')
						c=va_arg(va, int);
					else
						break;
				n*=10;
				n+=c;
			}
		}
		lflag=FALSE;
		if (c=='l'||c=='L')
		{
			lflag++;
			if (*fmt)
				c=*fmt++;
		}
		switch (c)
		{
		case 'X':
			lflag=TRUE;
		case 'x':
			c=16;
			goto oxu;
		case 'U':
			lflag=TRUE;
		case 'u':
			c=10;
			goto oxu;
		case 'O':
			lflag=TRUE;
		case 'o':
			c=8;
			oxu: if (lflag)
				l=va_arg(va, long);
			else
				l=va_arg(va, int);
			p=ltoa(p,l,c);
			break;
		case 'D':
			lflag=TRUE;
		case 'd':
			if (lflag)
				l=va_arg(va, long);
			else
				l=va_arg(va, int);
			if (l<0)
			{
				*p++='-';
				l=-l;
			}
			p=ltoa(p,l,10);
			break;
		case 'c':
			filler=' ';
			*p++=va_arg(va, int);
			break;
		case 's':
			filler=' ';
			if ((s=va_arg(va, char *))==0)
				s="(null)";
			if (n==0)
				n=32767;
			for (p=s; *p&&--n>=0; p++)
				;
			break;
		default:
			*p++=c;
			break;
		}
		i=(int)(p-s);
		if ((width-=i)<0)
			width=0;
		if (ljust==FALSE)
			width=-width;
		if (width<0)
		{
			if (*s=='-'&&filler=='0')
			{
				if(dest)
					*dest++=*s++;
				else
					printfch(*s++);
				i--;
			}
			do
				if(dest)
					*dest++=filler;
				else
					printfch(filler);
			while (++width!=0);
		}
		while (--i>=0)
		{
			if(dest)
				*dest++=*s++;
			else
				printfch(*s++);
		}

		while (width)
		{
			if(dest)
				*dest++=filler;
			else
				printfch(filler);
			width--;
		}
	}
}

void chsprintf(char *dest, const char *fmt, ...)
{
	va_list va;

	va_start(va,fmt);
	chvsprintf(dest,fmt,va,0);
	va_end(va);
}

//who put u32 and u16 as structure variable names geez
#undef u32
#undef u16
#undef u8
static void swoWrite(int channel, uint8_t *data, int length)
{
	if (CoreDebug->DEMCR&CoreDebug_DEMCR_TRCENA_Msk) // Trace enabled
	{
		if (ITM->TCR&ITM_TCR_ITMENA_Msk) // ITM enabled
		{
			if (ITM->TER&(1ul<<channel)) // ITM Port #0 enabled
			{
				while (length)
				{
					while (ITM->PORT[channel].u32==0)
						;
					if(length>=4)
					{
						ITM->PORT[channel].u32=(data[3]<<24)|(data[2]<<16)|(data[1]<<8)|data[0];
						data+=4;
						length-=4;
					}else
						if(length>=2)
						{
							ITM->PORT[channel].u16=(data[1]<<8)|data[0];
							data+=2;
							length-=2;
						}else
						{
							ITM->PORT[channel].u8=*data;
							data++;
							length--;
						}
				}
			}
		}
	}
}
